
/**
 * Methods to verify and print exam schedules
 */

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.Map;

public class ScheduleChecker {
    public static void print(Iterable<? extends Collection<String>> schedule) {
        int slotCount = 0;
        for (Collection<String> slot : schedule) {
            slotCount++;
            System.out.print("Slot " + slotCount + ": ");
            String[] slotArr = slot.toArray(new String[0]);
            for (int i = 0; i < slotArr.length - 1; i++) {
                System.out.print(slotArr[i] + ", ");
            }
            System.out.println(slotArr[slotArr.length - 1]);
        }
    }

    public static void check(UndirectedGraph g, Iterable<? extends Collection<String>> schedule) {
        System.out.println(
                "Checking schedule against graph with " + g.numVertices() + " vertices and " + g.numEdges() + " edges");
        print(schedule);
        // check that every vertex is assigned to a slot
        // check that no vertex is assigned to multiple slots
        // check that every scheduled course is in the graph
        int errorCount = 0;
        int slotCount = 0;
        Map<String, Integer> counter = new HashMap<>();
        for (Collection<String> slot : schedule) {
            slotCount++;
            for (String course : slot) {
                if (Collections.frequency(slot, course) > 1) {
                    System.out.println("ERROR: " + course + " appears multiple times in the same slot: " + slot);
                    errorCount++;
                }
                if (!g.containsVertex(course)) {
                    System.out.println("ERROR: " + course + " appears in schedule, but not in the graph");
                    errorCount++;
                }
                counter.put(course, counter.getOrDefault(course, 0) + 1);
            }
        }
        System.out.println("Schedule consists of " + slotCount + " slots and " + counter.size() + " courses");
        for (String vertex : g.V()) {
            if (!counter.containsKey(vertex)) {
                System.out.println("ERROR: " + vertex + " appears in the graph, but not the schedule");
                errorCount++;
            } else if (counter.get(vertex) > 1) {
                System.out.println("ERROR: " + vertex + " is scheduled multiple times");
                errorCount++;
            }
        }

        if (errorCount == 0) {
            System.out.println("SUCCESS: schedule includes exactly the courses from the graph exactly once");
        }

        // check for conflicts
        for (Collection<String> slot : schedule) {
            String[] slotArr = slot.toArray(new String[0]);
            for (int i = 0; i < slot.size(); i++) {
                for (int j = i + 1; j < slot.size(); j++) {
                    if (g.containsEdge(slotArr[i], slotArr[j])) {
                        System.out.println("ERROR: " + slotArr[i] + " and " + slotArr[j]
                                + " are scheduled at the same time, but have a conflict according to the graph");
                        errorCount++;
                    }
                }
            }
        }

        // check whether any two slots can be merged
        for (Collection<String> slot : schedule) {
            for (Collection<String> otherSlot : schedule) {
                if (slot != otherSlot) {
                    ArrayList<String> combinedSlot = new ArrayList<>(slot);
                    combinedSlot.addAll(otherSlot);
                    boolean conflict = false;
                    for (int i = 0; i < combinedSlot.size(); i++) {
                        for (int j = i + 1; j < combinedSlot.size(); j++) {
                            conflict = conflict || g.containsEdge(combinedSlot.get(i), combinedSlot.get(j));
                        }
                    }
                    if (!conflict) {
                        System.out.println("ERROR: " + slot + " and " + otherSlot
                                + " can be combined without causing any conflicts");
                        errorCount++;
                    }
                }
            }
        }

        if (errorCount == 0) {
            System.out.println("SUCCESS: SCHEDULE VERIFIED, 0 errors");
        } else {
            System.out.println("ERROR: SCHEDULE INVALID, " + errorCount + " errors");
        }
    }
}
